import pandas as pd
import numpy as np
import statsmodels.api as sm
from enum import Enum
import os

class TexRessources:
    m_tex_1 = "% $Id: jfsample.tex,v 19:a118fd22993e 2013/05/24 04:57:55 stanton $" + "\n" + \
              "\documentclass[12pt]{article}" + "\n" + \
              r"" + "\n" + \
              r"% DEFAULT PACKAGE SETUP" + "\n" + \
              r"" + "\n" + \
              r"\usepackage{setspace,graphicx,epstopdf,amsmath,amsfonts,amssymb,amsthm,versionPO}" + "\n" + \
              r"\usepackage{marginnote,datetime,enumitem,subfigure,rotating,fancyvrb}" + "\n" + \
              r"%\usepackage{marginnote,datetime,enumitem,subfigure,fancyvrb}" + "\n" + \
              r"\usepackage{float}" + "\n" + \
              r"\usepackage[" + "\n" + \
              r"  colorlinks=true," + "\n" + \
              r"  linkcolor=blue," + "\n" + \
              r"  citecolor=blue]{hyperref}" + "\n" + \
              r"\usepackage{booktabs}" + "\n" + \
              r"\usepackage[longnamesfirst]{natbib}" + "\n" + \
              r"\usepackage{mathtools}" + "\n" + \
              r"\usepackage{lscape}" + "\n" + \
              r"\usepackage[toc,page]{appendix}" + "\n" + \
              r"% custom antoine stuff" + "\n" + \
              r"\def\rot{\rotatebox}" + "\n" + \
              r"\usepackage{tikz}" + "\n" + \
              r"\usetikzlibrary{positioning}" + "\n" + \
              r"\usepackage{forest}" + "\n" + \
              r"\newtheorem{prop}{Proposition}" + "\n" + \
              r"\usetikzlibrary{trees}" + "\n" + \
              r"" + "\n" + \
              r"\usdate" + "\n" + \
              r"" + "\n" + \
              r"% These next lines allow including or excluding different versions of text" + "\n" + \
              r"% using versionPO.sty" + "\n" + \
              r"" + "\n" + \
              r"\excludeversion{notes}        % Include notes?" + "\n" + \
              r"\includeversion{links}          % Turn hyperlinks on?" + "\n" + \
              r"" + "\n" + \
              r"% Turn off hyperlinking if links is excluded" + "\n" + \
              r"\iflinks{}{\hypersetup{draft=true}}" + "\n" + \
              r"" + "\n" + \
              r"% Notes options" + "\n" + \
              r"\ifnotes{%" + "\n" + \
              r"\usepackage[margin=1.0in,paperwidth=10in,right=2.5in]{geometry}%" + "\n" + \
              r"\usepackage[textwidth=1.4in,shadow,colorinlistoftodos]{todonotes}%" + "\n" + \
              r"}{%" + "\n" + \
              r"\usepackage[margin=1in]{geometry}%" + "\n" + \
              r"\usepackage[disable]{todonotes}%" + "\n" + \
              r"}" + "\n" + \
              r"" + "\n" + \
              r"\newlength\mytemplength" + "\n" + \
              r"\newcommand\parboxc[3]{%" + "\n" + \
              r"    \settowidth{\mytemplength}{#3}%" + "\n" + \
              r"    \parbox[#1][#2]{\mytemplength}{\centering #3}%" + "\n" + \
              r"}" + "\n" + \
              r"" + "\n" + \
              r"% make footnote pripritary so we don't split them from page to page" + "\n" + \
              r"\interfootnotelinepenalty=10000" + "\n" + \
              r"% Allow todonotes inside footnotes without blowing up LaTeX" + "\n" + \
              r"% Next command works but now notes can overlap. Instead, we'll define" + "\n" + \
              r"% a special footnote note command that performs this redefinition." + "\n" + \
              r"%\renewcommand{\marginpar}{\marginnote}%" + "\n" + \
              r"" + "\n" + \
              r"% Save original definition of \marginpar" + "\n" + \
              r"\let\oldmarginpar\marginpar" + "\n" + \
              r"" + "\n" + \
              r"% Workaround for todonotes problem with natbib (To Do list title comes out wrong)" + "\n" + \
              r"\makeatletter\let\chapter\@undefined\makeatother % Undefine \chapter for todonotes" + "\n" + \
              r"" + "\n" + \
              r"% Define note commands" + "\n" + \
              r"\newcommand{\smalltodo}[2][] {\todo[caption={#2}, size=\scriptsize, fancyline, #1] {\begin{spacing}{0.5}#2\end{spacing}}}" + "\n" + \
              r"\newcommand{\rhs}[2][]{\smalltodo[color=green!30,#1]{{\bf RS:} #2}}" + "\n" + \
              r"\newcommand{\rhsnolist}[2][]{\smalltodo[nolist,color=green!30,#1]{{\bf RS:} #2}}" + "\n" + \
              r"\newcommand{\rhsfn}[2][]{%  To be used in footnotes (and in floats)" + "\n" + \
              r"\renewcommand{\marginpar}{\marginnote}%" + "\n" + \
              r"\smalltodo[color=green!30,#1]{{\bf RS:} #2}%" + "\n" + \
              r"\renewcommand{\marginpar}{\oldmarginpar}}" + "\n" + \
              r"%\newcommand{\textnote}[1]{\ifnotes{{\noindent\color{red}#1}}{}}∫" + "\n" + \
              r"\newcommand{\textnote}[1]{\ifnotes{{\colorbox{yellow}{{\color{red}#1}}}}{}}" + "\n" + \
              r"\newcommand{\fixmeAntoine}[1]{\textcolor{red}{\textbf{TBD Antoine: #1}}}" + "\n" + \
              r"\newcommand{\fixmeHui}[1]{\textcolor{blue}{\textbf{TBD Hui: #1}}}" + "\n" + \
              r"\newcommand{\fixmeSimon}[1]{\textcolor{purple}{\textbf{TBD Simon: #1}}}" + "\n" + \
              r"\newcommand{\alll}[1]{\textcolor{teal}{\textbf{TBD All: #1}}}" + "\n" + \
              r"" + "\n" + \
              r"% \input{tex/commands.tex}" + "\n" + \
              r"" + "\n" + \
              r"% Command to start a new page, starting on odd-numbered page if twoside option" + "\n" + \
              r"% is selected above" + "\n" + \
              r"\newcommand{\clearRHS}{\clearpage\thispagestyle{empty}\cleardoublepage\thispagestyle{plain}}" + "\n" + \
              r"" + "\n" + \
              r"% Number paragraphs and subparagraphs and include them in TOC" + "\n" + \
              r"\setcounter{tocdepth}{2}" + "\n" + \
              r"\usepackage{bbm}" + "\n" + \
              r"% JF-specific includes:" + "\n" + \
              r"" + "\n" + \
              r"\usepackage{indentfirst} % Indent first sentence of a new section." + "\n" + \
              r"%\usepackage{endnotes}    % Use endnotes instead of footnotes" + "\n" + \
              r"\usepackage{jf}          % JF-specific formatting of sections, etc." + "\n" + \
              r"\usepackage[labelfont=bf,labelsep=period]{caption}   % Format figure captions" + "\n" + \
              r"\captionsetup[table]{labelsep=none}" + "\n" + \
              r"" + "\n" + \
              r"% Define theorem-like commands and a few random function names." + "\n" + \
              r"\newtheorem{condition}{CONDITION}" + "\n" + \
              r"\newtheorem{corollary}{COROLLARY}" + "\n" + \
              r"\newtheorem{proposition}{PROPOSITION}" + "\n" + \
              r"\newtheorem{obs}{OBSERVATION}" + "\n" + \
              r"\newcommand{\argmax}{\mathop{\rm arg\,max}}" + "\n" + \
              r"\newcommand{\defeq}{\stackrel{\rm def}{=}}" + "\n" + \
              r"\def\QQ{{\mathbb Q}}" + "\n" + \
              r"" + "\n" + \
              r"\usepackage{caption}" + "\n" + \
              r"" + "\n" + \
              r"" + "\n" + \
              r"\captionsetup[table]{" + "\n" + \
              r"  labelsep=newline," + "\n" + \
              r"  justification=centerfirst" + "\n" + \
              r"}" + "\n" + \
              r"" + "\n" + \
              r"\usepackage{changes}" + "\n" + \
              r"\definechangesauthor[name={HC}, color=red]{HC}" + "\n" + \
              r"" + "\n" + \
              r"" + "\n" + \
              r"" + "\n" + \
              r"\begin{document}" + "\n" + \
              r"" + "\n" + \
              r"\setlist{noitemsep}  % Reduce space between list items (itemize, enumerate, etc.)" + "\n" + \
              r"\onehalfspacing      % Use 1.5 spacing" + "\n" + \
              r"%\doublespacing      % Use 2.0 spacing" + "\n" + \
              r"\parskip 3pt" + "\n" + \
              r"" + "\n" + \
              r"" + "\n" + \
              r"" + "\n" + \
              r"\title{" + "\n"

    m_tex_2 = "}\n \\author{"

    m_tex_3 = "}" + "\n" + \
              r"" + "\n" + \
              r"\date{\today}" + "\n" + \
              r"" + "\n" + \
              r"% Create title page with no page number" + "\n" + \
              r"" + "\n" + \
              r"\maketitle" + "\n" + \
              r"\thispagestyle{empty}" + "\n" + \
              r"" + "\n" + \
              r"\bigskip" + "\n" + \
              r"" + "\n" + \
              r"\begin{abstract}" + "\n" + \
              r"\input{sec/abstract.tex}" + "\n" + \
              r"\end{abstract}" + "\n" + \
              r"" + "\n" + \
              r"\medskip" + "\n" + \
              r"" + "\n" + \
              r"%\begin{centering}" + "\n" + \
              r"%{\large{\textbf{This is a preliminary draft. Please do not cite or distribute without permission" + "\n" + \
              r"%of the authors.}}}\\" + "\n" + \
              r"%\end{centering}" + "\n" + \
              r"" + "\n" + \
              r"\clearpage" + "\n" + \
              r"" + "\n" + \
              r"\setcounter{page}{1}"
    m_tex_4 = r"% Bibliography." + "\n" + \
              "\n" + \
              r"\begin{onehalfspacing}   % Double-space the bibliography" + "\n" + \
              r"\bibliographystyle{jf}" + "\n" + \
              r"\bibliography{"
    m_tex_5 = r"}" + "\n" + \
              r"\end{onehalfspacing}" + "\n" + \
              r"\end{document}"

    versionPO_sty = r"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + "\n" + \
                    r"% You need Rainer Schoepf's verbatim.sty to use this!!" + "\n" + \
                    r"%" + "\n" + \
                    r"% Version control macros. These let you define environments whose contents" + "\n" + \
                    r"% will be optionally added to or deleted from the text when you run LaTeX." + "\n" + \
                    r"% Usage: place either of the following near the start of your file:" + "\n" + \
                    r"%     \includeversion{NAME}" + "\n" + \
                    r"%     \excludeversion{NAME}" + "\n" + \
                    r"% Here, NAME is any name you choose. The first one indicates that text" + "\n" + \
                    r"% between \begin{NAME} and \end{NAME} will be processed in the normal way." + "\n" + \
                    r"% The second indicates that text between \begin{NAME} and \end{NAME} will" + "\n" + \
                    r"% be totally deleted." + "\n" + \
                    r"% These command automatically declare the inverted version not-NAME." + "\n" + \
                    r"%" + "\n" + \
                    r"% NOTE:" + "\n" + \
                    r"% You cannot use the version environments inside another environment," + "\n" + \
                    r"% inside a command or in an argument to another command. You can make a" + "\n" + \
                    r"% limited use of a version inside an environment by using \NAME and" + "\n" + \
                    r"% \endNAME rather than \begin{NAME} and \end{NAME}. E.g." + "\n" + \
                    r"%   \newenvironment{example}{...\myversion...}{\endmyversion} where the" + "\n" + \
                    r"% following would fail:" + "\n" + \
                    r"%   \newenvironment{example}{...\begin{myversion}...}{\end{myversion}}" + "\n" + \
                    r"%" + "\n" + \
                    r"% To alleviate these problems a bit, foreach version NAME also a command" + "\n" + \
                    r"% \ifNAME is defined such that \ifNAME{yes-part}{no-part} selects either" + "\n" + \
                    r"% yes-part or no-part depending on whether the version is included or" + "\n" + \
                    r"% excluded. The \ifNAME command may be used everywhere but the environments" + "\n" + \
                    r"% are preferred with large amounts of text. Also the parameters of \ifNAME" + "\n" + \
                    r"% are not included in a group, so you can make global declarations in them." + "\n" + \
                    r"% The \ifNAME command is fragile if either of its arguments is fragile." + "\n" + \
                    r"% Note: the \ifNAME command is only usable if NAME consists of letters only." + "\n" + \
                    r"%" + "\n" + \
                    r"% A ``comment'' environment has already been pre-defined for you with " + "\n" + \
                    r"% \excludeversion{comment}; you can NOT override this one." + "\n" + \
                    r"%" + "\n" + \
                    r"% You can define environments for as many versions as you want. " + "\n" + \
                    r"%" + "\n" + \
                    r"% Example: " + "\n" + \
                    r"%   \includeversion{abridged}" + "\n" + \
                    r"%   \ifabridged{\large}{\small}" + "\n" + \
                    r"%   Text for the" + "\n" + \
                    r"%   \begin{abridged} " + "\n" + \
                    r"%      short" + "\n" + \
                    r"%   \end{abridged}" + "\n" + \
                    r"%   \begin{not-abridged}" + "\n" + \
                    r"%      long and really longwinded, opaque and boring" + "\n" + \
                    r"%   \end{not-abridged}" + "\n" + \
                    r"%   version of the paper. Punctuation works correctly\begin{not-abridged}" + "\n" + \
                    r"%   because sphack is used\end{not-abridged}." + "\n" + \
                    r"%   \begin{comment} This is deleted by default. \end{comment}" + "\n" + \
                    r"%" + "\n" + \
                    r"%  Piet van Oostrum -- <piet@cs.ruu.nl>" + "\n" + \
                    r"%  Idea based on a style file by Stephen Bellantoni, implementation based" + "\n" + \
                    r"%  on verbatim.sty by Rainer Schoepf <SCHOEPF@SC.ZIB-Berlin.DE>." + "\n" + \
                    r"" + "\n" + \
                    r"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + "\n" + \
                    r"" + "\n" + \
                    r"\@ifundefined{verbatim@@@}{\input{verbatim.sty}}{}" + "\n" + \
                    r"" + "\n" + \
                    r"\def\includeversion#1{%" + "\n" + \
                    r"  \expandafter\def\csname #1\endcsname{}%" + "\n" + \
                    r"  \expandafter\def\csname end#1\endcsname{}%" + "\n" + \
                    r"  \expandafter\let\csname not-#1\endcsname\comment" + "\n" + \
                    r"  \expandafter\let\csname endnot-#1\endcsname\endcomment" + "\n" + \
                    r"  \expandafter\def\csname if#1\endcsname##1##2{##1}" + "\n" + \
                    r"}" + "\n" + \
                    r"\def\excludeversion#1{%" + "\n" + \
                    r"  \expandafter\let\csname #1\endcsname\comment" + "\n" + \
                    r"  \expandafter\let\csname end#1\endcsname\endcomment" + "\n" + \
                    r"  \expandafter\def\csname not-#1\endcsname{}%" + "\n" + \
                    r"  \expandafter\def\csname endnot-#1\endcsname{}%" + "\n" + \
                    r"  \expandafter\def\csname if#1\endcsname##1##2{##2}" + "\n" + \
                    r"}"

    jf_sty = "% Formatting for Journal of Finance papers" + "\n" + \
             r"% Theorems, etc. (need to load amsthm first)" + "\n" + \
             r"\newtheoremstyle{jf}% name" + "\n" + \
             r"{6pt}% Space above" + "\n" + \
             r"{6pt}% Space below" + "\n" + \
             r"{\itshape}% Body font" + "\n" + \
             r"{}% Indent amount" + "\n" + \
             r"{}% Theorem head font" + "\n" + \
             r"{:}% Punctuation after theorem head" + "\n" + \
             r"{.5em}% Space after theorem head" + "\n" + \
             r"{}% Theorem head spec (can be left empty, meaning normal)" + "\n" + \
             r"" + "\n" + \
             r"\theoremstyle{jf}" + "\n" + \
             r"" + "\n" + \
             r"% Make References title all caps." + "\n" + \
             r"\renewcommand{\refname}{REFERENCES}" + "\n" + \
             r"" + "\n" + \
             r"% Put dots after numbers in section headers." + "\n" + \
             r"\renewcommand{\@seccntformat}[1]{{\csname the#1\endcsname}.\hspace{1em}}" + "\n" + \
             r"" + "\n" + \
             r"% Section and table numbers " + "\n" + \
             r"\def\thesection       {\Roman{section}}" + "\n" + \
             r"\def\thesubsection    {\Alph{subsection}}" + "\n" + \
             r"\def\thetable         {\Roman{table}}" + "\n" + \
             r"" + "\n" + \
             r"% Section header format" + "\n" + \
             r"\renewcommand{\section}{\@startsection" + "\n" + \
             r"  {section}{1}{0mm}{-3.5ex \@plus -1ex \@minus -.2ex}{2.3ex \@plus.2ex}{\centering\normalfont\Large\bfseries}}" + "\n" + \
             r"" + "\n" + \
             r"\renewcommand{\subsection}{\@startsection" + "\n" + \
             r"  {subsection}{2}{0mm}{-3.25ex \@plus -1ex \@minus -.2ex}{1.5ex \@plus.2ex}{\normalfont\large\itshape}}" + "\n" + \
             r"" + "\n" + \
             r"% Put section number back in references to subsections and subsubsections." + "\n" + \
             r"\renewcommand{\p@subsection}{\thesection .}" + "\n" + \
             r"\renewcommand{\p@subsubsection}{\thesection .}" + "\n" + \
             r"" + "\n" + \
             r"% Appendix formatting:" + "\n" + \
             r"" + "\n" + \
             r"\def\appendix{\par" + "\n" + \
             r"  \setcounter{section}{0}%         % Start counting sections again" + "\n" + \
             r"  \setcounter{subsection}{0}% " + "\n" + \
             r"% Use uppercase letters for section numbers:" + "\n" + \
             r"  \gdef\thesection{\@Alph\c@section}% " + "\n" + \
             r"% Number equations (A1), etc.:" + "\n" + \
             r"  \renewcommand{\theequation}{\thesection\arabic{equation}}% " + "\n" + \
             r'% Put word "Appendix" before appendix number:' + "\n" + \
             r"  \renewcommand{\@seccntformat}[1]{{Appendix \csname the##1\endcsname}.\hspace{1em}}" + "\n" + \
             r"% Same formatting as above, but reset equation counter with each section" + "\n" + \
             r"  \renewcommand{\section}{\setcounter{equation}{0}\@startsection" + "\n" + \
             r"    {section}{1}{0mm}{-3.5ex \@plus -1ex \@minus -.2ex}{2.3ex \@plus.2ex}{\centering\normalfont\Large\bfseries}}" + "\n" + \
             r"}" + "\n" + \
             r"" + "\n"



import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.metrics import r2_score

class OlsPLus:
    def __init__(self):
        pass

    def sm_with_fixed_effect(self,df,y,x,fix_effect_cols=[], std_error_cluster=None):
        temp = df[[y]+x+fix_effect_cols].copy().dropna()
        temp['y_untr'] = temp[y]
        t = 0
        if len(fix_effect_cols)>0:
            for fe in fix_effect_cols:
                t += temp.groupby(fe)[y].transform('mean')
                temp[y] = temp[y]-t
                for c in x:
                    temp[c] = temp[c]-temp.groupby(fe)[c].transform('mean')

        if std_error_cluster is None:
            m=sm.OLS(temp[y],temp[x]).fit()
        else:
            m=sm.OLS(temp[y],temp[x]).fit(cov_type='cluster',cov_kwds={'groups':temp[std_error_cluster]})


        if len(fix_effect_cols)>0:
            pred=m.predict(temp[x])+t
            r = r2_score(temp['y_untr'],pred)
            m.rsquared = r
            m.rsquared_adj = 1-(1-r)*(temp.shape[0]-1)/(temp.shape[0]-temp.shape[1]-1)
        return m


import pandas as pd
import numpy as np
import statsmodels.api as sm
from enum import Enum

class ParValue(Enum):
    TSTAT = 1
    PVALUE = 2
    STD = 3

class OneReg:
    def __init__(self, reg, show_list=[], hide_list=[], blocks=[], bottom_blocks=[]):
        self.reg = reg
        if show_list == []:
            self.show_list = []
            for s in self.reg.params.keys():
                if s not in hide_list:
                    self.show_list.append(s)
        else:
            self.show_list = show_list

        self.blocks = blocks
        self.bottom_block = bottom_blocks

    def create_columns(self):
        # first add the parameters of the reg
        d = pd.Series(dtype=object)
        for k in self.show_list:
            if k in self.reg.params.keys():
                v = self.reg.pvalues[k]
                p = f'{np.round(self.reg.params[k], TableReg.round):,}'
                for tr in TableReg.sign_tr:
                    if v <= tr:
                        p += '*'
                # self = table_2.reg_list[0]
                # update the v to be tstat or std depending on parameters
                if TableReg.par_value == ParValue.TSTAT:
                    v = self.reg.tvalues[k]
                if TableReg.par_value == ParValue.STD:
                    v = self.reg.bse[k]
                v = r'(' + f'{np.round(v, TableReg.round):,}' + r')'
                v_l = [len(x) for x in v.split('.')]
                p_l = [len(x) for x in p.split('.')]

                t = abs(v_l[0] - p_l[0])
                t = r'\phantom{' + '*' * t + '}'
                if v_l[0] > p_l[0]:
                    p = t + p
                if v_l[0] < p_l[0]:
                    v = t + v
                t = abs(v_l[1] - p_l[1])
                t = r'\phantom{' + '*' * t + '}'
                if v_l[1] > p_l[1]:
                    p = p+t
                if v_l[1] < p_l[1]:
                    v = v+t
                # else:
                    # p = r'\phantom{(}' + p + r'\phantom{)}'
                d[k] = p

                t = pd.Series(dtype=object)
                t[''] =v

                d = pd.concat([d,t],axis=0)
                
            else:
                t = pd.Series(dtype=object)
                t[k] = TableReg.missing_symbol
                t[''] = TableReg.missing_symbol
                d = pd.concat([d,t],axis=0)
        # now we can add the "blocks", that is fix effects and others
        for block in self.blocks:
            t = pd.Series(dtype=object)
            t[TableReg.group_key] = ''

            for k in block.keys():
                t[k] = block[k]
            d = pd.concat([d,t],axis=0)

        # finaly additional info (r² and n.obs per default, but you can add anything through bottom blocks
        if TableReg.show_obs | TableReg.show_r2 | (len(self.bottom_block)>0):
            t = pd.Series(dtype=object)
            t[TableReg.group_key] = ''
            t['Observations'] = f'{int(self.reg.nobs):,}'

            if hasattr(self.reg,'rsquared_adj'):

                t[r'$R^2$'] = np.round(self.reg.rsquared_adj,TableReg.round_r2)
            else:
                t[r'Pseudo $R^2$'] = np.round(self.reg.prsquared,TableReg.round_r2)

            first_block = True
            for block in self.bottom_block:
                if first_block:
                    first_block = False
                else:
                    t[TableReg.group_key] = ''
                t = pd.Series(dtype=object)
                for k in block.keys():
                    t[k] = block[k]
            d = pd.concat([d,t],axis=0)
        return d


class TableReg:
    missing_symbol = ' '
    par_value = ParValue.STD
    round = 4
    round_r2 = 4
    sign_tr = [0.1, 0.05, 0.01]
    show_obs = True
    show_r2 = True
    variable_skip = r'\smallskip'
    group_key = 'asgeg'
    group_skip = r'\medskip'
    equal_lines = False

    def __init__(self, **option):
        self.reg_list = []
        self.hide_list = []
        self.order = []
        self.df = None
        self.final_show_list = []
        self.show_only_list = []
        self.col_groups = []
        self.rename_dict = {}
        if 'hide_list' in option:
            assert type(option['hide_list']) == list, "The overall hide list has to be a list"
            self.hide_list = option['hide_list']

        if 'show_only_list' in option:
            assert type(option['show_only_list']) == list, "The show only  list has to be a list"
            self.show_only_list = option['show_only_list']

        if 'order' in option:
            assert type(option['order']) == list, "The order has to be a list"
            self.order = option['order']

        if 'col_groups' in option:
            self.set_col_groups(option['col_groups'])

        if 'rename_dict' in option:
            self.set_rename_dict(option['rename_dict'])


    def set_rename_dict(self, rename_dict):
        assert type(rename_dict) == dict, "The rename dict must be a dictionary"
        self.rename_dict = rename_dict



    def set_col_groups(self, groups):
        assert type(groups) == list, "The col order has to be a list of list"
        for group in groups:
            assert type(group) == list, "Each col group must be a list ['name of group', first columne in the group (int), last col in group (int)]"
        self.col_groups = groups


    def add_reg(self, reg, show_list=[], hide_list=[], blocks=[],bottom_blocks=[]):
        hide_list = hide_list + self.hide_list
        self.reg_list.append(OneReg(reg, show_list, hide_list, blocks, bottom_blocks))

    def update_show_list(self):
        if len(self.show_only_list) == 0:
            show_list = []
            for oneReg in self.reg_list:
                show_list = list(set(show_list + oneReg.show_list))
            show_list = list(np.sort(show_list))
            show_list = self.order + [x for x in show_list if x not in self.order]
        else:
            show_list = self.show_only_list

        col = []
        for oneReg in self.reg_list:
            oneReg.show_list = show_list
            col.append(oneReg.create_columns())

        self.df = pd.concat(col,axis=1)

        self.df.columns = [r'\parboxc{c}{0.6cm}{('+str(int(i+1))+')}' for i in range(self.df.shape[1])]
        self.df = self.df.rename(index=self.rename_dict)

        self.final_show_list = show_list
        self.final_show_list = pd.Series(self.final_show_list).replace(self.rename_dict).values.tolist()
        self.tex=''

    def create_tex(self):
        self.update_show_list()

        # writing the tex modification to include name templatess
        tex = self.df.to_latex(escape=False)
        cols = tex.split('\\begin{tabular}{')[1].split('}')[0]
        rep = list(cols.replace('l','c'))
        rep[0] = 'l'
        tex = tex.replace(cols,''.join(rep))

        if len(self.col_groups)>0:
            # adding "group col names"
            s = '\n '
            s_line = '\n '
            for g in self.col_groups:
                s += '& \multicolumn{'+str(1+g[2]-g[1])+'}{c}{\parboxc{c}{0.6cm}{'+g[0]+'}}'
                # s += '& \multicolumn{'+str(1+g[2]-g[1])+'}{c}{'+g[0]+'}'
                s_line += r'\cmidrule(lr){'+str(g[1]+1)+'-'+str(g[2]+1)+'}'
            s += r' \\'+'\n'
            s_line += '\n'

            ts = tex.split(r'\toprule')
            tex = ts[0]+r'\toprule' + s +s_line+ ts[1]

            ts = tex.split(r'\midrule')
            tex = ts[0]+r'\midrule' + ts[1]

        # adding the skip between variable
        # first we extract the maxium length of a column on the first one
        L = 0
        for x in self.df.index:
            L = max(L,len(x))
        L+=1
        for i in range(1,len(self.final_show_list)):
            a = self.final_show_list[i]
            # a += ' '*(L-len(a))+'&'
            a += ' &'
            ts = tex.split(a)
            temp = ts[0][:-4] + TableReg.variable_skip + ts[0][-4:]
            tex=temp+a+ts[1]


        # processing the group skip
        t = None
        for item in tex.split("\n"):
            if TableReg.group_key in item:
                t = item
        # replacing specific rule

        if t is not None:
            self.tex = tex.replace(t, TableReg.group_skip + r'\\')
        else:
            self.tex = tex

    def save_tex(self, save_dir):

        self.create_tex()
        tex = self.tex
        if TableReg.equal_lines:
            tex=tex.replace(r'\toprule',r'\hline')
            tex=tex.replace(r'\midrule',r'\hline')
            tex=tex.replace(r'\bottomrule',r'\hline')

        with open(save_dir,'w') as txt:
            txt.write(tex)

    @staticmethod
    def create_panel_of_tables(table_list, name_list, save_dir):
        numbers = 'A B C D E F G H I J K L M N O P Q R S T U V W X Y Z'.split()
        title_list = []
        for i in range(len(table_list)):
            table_list[i].create_tex()
            title_list.append('Panel '+numbers[i]+': '+name_list[i])


        tex = table_list[0].tex
        temp = r' \multicolumn{6}{c}{\parboxc{c}{0.7cm}{'+title_list[i]+r'}} \\'
        ts = tex.split(r'\toprule')
        tex = ts[0]+r'\toprule' +temp+r'\hline'+ts[1]
        tex = tex.replace(r'\bottomrule','')
        tex = tex.replace(r'\end{tabular}',r'asf')
        tex = tex.replace('\\\\\n\nasf','\\bigskip \\\\ \n')


        for i in range(1,len(table_list)):
            t_tex = table_list[i].tex
            temp = r' \multicolumn{6}{c}{\parboxc{c}{0.6cm}{' + title_list[i] + r'}} \\'
            ts = t_tex.split(r'\toprule')
            t_tex = ts[0] + r'\hline' + temp + r'\hline' + ts[1]
            t = None
            for item in t_tex.split("\n"):
                if r'\begin{tabular}' in item:
                    t = item
            t_tex = t_tex.replace(t,'')
            if i+1 < len(table_list):
                t_tex = t_tex.replace(r'\bottomrule','')
                t_tex = t_tex.replace(r'\end{tabular}', r'asf')
                t_tex = t_tex.replace('\\\\\n\nasf', '\\bigskip \\\\ \n')
            tex +=t_tex



        with open(save_dir,'w') as txt:
            txt.write(tex)






import tqdm
from didipack.create_rf.RandomFeaturesGenerator import RandomFeaturesGenerator
import os
import sys
import pandas as pd
import pickle
import numpy as np
from didipack.create_rf.aux import TicToc, get_block_sizes

def generate_random_features_in_blocks(
        x_mat: np.ndarray,
        y_mat: np.ndarray,
        date_ids: np.ndarray,
        nb_of_random_features_total: int,
        block_size_for_generation: int,
        save_dir: str,
        random_features_generator: RandomFeaturesGenerator,
        start_seed: int =0,
        voc_grid: list = None,
        para_nb_of_list_group:int = None,
        para_id: None = int,
        save_results = True
):
    '''
    Generates either the full matrix for rf (if para_grid is none) or a large chunk of it block by block.
    Save it in the save_dir to be merged latter.
    :param x_mat: input of raw signals (N*T, p)
    :param y_mat: input of corresponding returns (N*T,1)
    :param date_ids: input of dates for each individual row of x_mat (N*T,1)
    :param nb_of_random_features_total: the value of big P, total number of rf you want.
    :param block_size_for_generation: the maximum size of a block. used to generate the random feature block by block
    :param save_dir: the path to directory were to save intermediary results
    :param random_features_generator: RandomFeaturesGenerator object already initialized with the objects you want.
    :param start_seed: seed that correspond to the overall operation
    :param voc_grid: if you want to use a voc grid with some P in it smaller than the max block size, you shuold include a list of it to make sure each chunk is of the right size.

    The next two parameters concerns paralelization in a server
    :param para_nb_of_list_group: if none, we run the full list of block in one go. Otherwise, we split the list in para_nb_of_list_group chunks that can be run in paralel.
    :param para_id: if para_nb_of_list_group is not none, this will define which of the group of the block list should be run. You should iterate over all of this to run the full P.

    :param redo_already_computed_block: if false, we only run the batch if it wasn't already save.
    :return: None
    '''

    # To run all the blocks and have the correct number of features add the end, we need to add one block size generation here to the total number of features we aim for
    # see line to_run = np.arange(0,len(blocks)-1,1) and the -1 to understand why
    # nb_of_random_features_total += block_size_for_generation

    tictoc = TicToc()
    blocks = get_block_sizes(nb_of_random_features_total, block_size_for_generation, voc_grid=voc_grid)
    os.makedirs(save_dir, exist_ok=True)
    # the list of all individual blocks to run
    to_run = np.arange(0,len(blocks)-1,1)
    if para_nb_of_list_group is not None:
        # we will split the loop into max(para_grid+1) chunks, and run the para_id th one
        to_run = np.array_split(to_run, para_nb_of_list_group)[para_id]
        print('split into', len(to_run), 'lists',flush=True)

    rf_factors_all = []

    for i_block in tqdm.tqdm(to_run,'Loop of iblock'):
        rf_factors = random_features_generator.generate_random_features_from_list_with_potential_ranking(
            seed=int((start_seed + 1) * 1e3) + i_block,
            batch_nb_rf=blocks[i_block+1]-blocks[i_block],
            # it seems we are generate each number_features_in_subset for each gamma
            date_ids=date_ids,
            y_mat=y_mat,
            x_mat=x_mat,
        )
        rf_factors_all.append(rf_factors)


    rf_factors = pd.concat(rf_factors_all, axis=1)
    if save_results:
        if para_id is not None:
            rf_factors.to_pickle(save_dir+f'rf_chunk_{para_id}.p')
            tictoc.toc(f"Finished and saved 1/{para_id} of the chunks")
        else:
            rf_factors.to_pickle(save_dir+f'rf.p')
            tictoc.toc("Finished and saved")

    return rf_factors
#
#
# def generate_random_features_in_blocks_if_not_already_ran(
#         x_mat: np.ndarray,
#         y_mat: np.ndarray,
#         date_ids: np.ndarray,
#         nb_of_random_features_total: int,
#         block_size_for_generation: int,
#         save_dir: str,
#         random_features_generator: RandomFeaturesGenerator,
#         start_seed: int =0,
#         voc_grid: list = None,
#         para_nb_of_list_group:int = None,
#         para_id: None = int,
#         save_results = True
# ):
#     # correcting for the missing chunk
#     nb_of_random_features_total += block_size_for_generation
#
#     if os.path.exists(save_dir + f'rf_chunk_{para_id}.p'):
#         print(f'ALREADY RAN, skip that one {para_id}',flush=True)
#     else:
#         print(f'Start running {para_id}',flush=True)
#         generate_random_features_in_blocks(
#             x_mat = x_mat,
#             y_mat = y_mat,
#             date_ids = date_ids,
#             nb_of_random_features_total = nb_of_random_features_total,
#             block_size_for_generation = block_size_for_generation,
#             save_dir = save_dir,
#             random_features_generator = random_features_generator,
#             start_seed = start_seed,
#             voc_grid = voc_grid,
#             para_nb_of_list_group = para_nb_of_list_group,
#             para_id = para_id,
#             save_results=save_results
#         )


import time
import numpy as np


class TicToc:  # simple class to measure the time of various part of the code
    def __init__(self):
        self.start_time = 0

    def tic(self):
        self.start_time = time.time()

    def toc(self, message=''):
        print(f'{message}: in {np.round((time.time() - self.start_time) / 60, 2)}m', flush=True)
        self.tic()


def get_block_sizes(
        number_random_features: int, small_subset_size: int, voc_grid: list = None
) -> list:
    """returns a list of block sizes

    We wish to produce a grid of block that will be used to generate the random features in chunks.

    The largest size will be small_subset_size. In addittion, we will include all points in the voc_grid.

    """

    block_sizes = (
        np.arange(0, number_random_features, small_subset_size).astype(int).tolist()
    )
    if number_random_features not in block_sizes:
        block_sizes += [number_random_features]

    # if grid point in voc_grid is not in block_sizes, we add them in the block_sizes
    if voc_grid:
        block_sizes = list(set(block_sizes + voc_grid))
    block_sizes.sort()  # sort grid points
    return block_sizes







import numpy as np
import pandas as pd
import tqdm

class General:
    @staticmethod
    def split_array_fix_size(input_array, fix_size):
        return [input_array[i:i + fix_size] for i in range(0, len(input_array), fix_size)]

    @staticmethod
    def smart_chunks(lst, n):
        """Choose chunking method based on deviation from the desired chunk size."""
        chunk_count = len(lst) // n
        remainder = len(lst) % n

        # Calculate the size of the last chunk with the leftover method
        leftover_last_chunk_size = remainder if remainder != 0 else n

        # Calculate the size of the last chunk with the equally spread method
        spread_last_chunk_size = n + 1 if remainder > chunk_count else n

        # If the size of the last chunk with the leftover method is closer to n, use chunks_chunks_if_leftover_list
        if abs(n - leftover_last_chunk_size) <= abs(n - spread_last_chunk_size):
            return General.chunks_chunks_if_leftover_list(lst, n)

        # Otherwise, use chunks_equally_spread
        else:
            return General.chunks_equally_spread(lst, n)

    @staticmethod
    def chunks_chunks_if_leftover_list(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    @staticmethod
    def chunks_equally_spread(lst, n):
        """Yield successive n-sized chunks from lst, spreading remainder equally."""
        chunk_count = len(lst) // n
        remainder = len(lst) % n
        iterator = iter(lst)

        full_chunks = []
        for _ in range(chunk_count):
            full_chunks.append([next(iterator) for _ in range(n)])

        for i in range(remainder):
            full_chunks[i % len(full_chunks)].append(next(iterator))

        return full_chunks







import sys
import pandas as pd
from didipack.trainer.trainer_abstract import BaseTrainer
from didipack.utils_didi.ridge import run_efficient_ridge
import numpy as np
from didipack import Params


class TrainerRidge(BaseTrainer):
    def __init__(self, par: Params):
        super().__init__(par)
        self.beta = None

    def _predict(self, x):
        return x.values @ self.beta.T

    def _train_model(self, x: pd.DataFrame, y: pd.DataFrame, hyper_params):
        beta = run_efficient_ridge(signals=x.values, labels=y.values,shrinkage_list=hyper_params['shrinkage'])
        self.beta= beta

    def _validation_procedure(self, x_train: pd.DataFrame, y_train: pd.DataFrame, x_val: pd.DataFrame,
                              y_val: pd.DataFrame, hyper_parameters_set):
        self.beta = run_efficient_ridge(signals=x_train.values, labels=y_train.values,shrinkage_list=self.par.train.shrinkage_list)
        y_pred = self._predict(x_val)
        val_perf =np.mean(np.square(y_pred-y_val.values),axis=0)
        best_lambda = self.par.train.shrinkage_list[np.argmin(val_perf)]
        print('Select best lambda', best_lambda,flush=True)
        return {'shrinkage':[best_lambda]}



if __name__ == "__main__":
    try:
        grid_id = int(sys.argv[1])
        model_id = int(sys.argv[2])
        print('Running with args', grid_id, flush=True)
    except:
        print('Debug mode on local machine', flush=True)
        grid_id = 0
        model_id = 5

    par = Params()
    par.train.testing_window = 5
    self = TrainerRidge(par)

    N = 1000
    P = 100
    fake_data = pd.DataFrame(np.random.normal(size=(N,P)))
    y = pd.DataFrame(fake_data.mean(1))
    dates = pd.Series(np.arange(N))
    ids = dates.copy()

    y_test, _ = self.train_at_time_t(x=fake_data,y=y,ids=ids,times=dates,t_index_for_split=100)




import numpy as np
import itertools
import pandas as pd
import os


class RandomFeaturesParams:
    def __init__(self):
        self.max_rf = 50*1000
        self.gamma_list = [0.5,0.6,0.7,0.8,0.9,1.0]
        self.block_size_for_generation=1000
        self.start_seed=0
        self.voc_grid=[100, 200, 360, 500, 1000, 2000, 5000, 10000, 20000, 50000, 100000, 200000, 500000]
        self.para_nb_of_list_group=20
        self.para_id=0


class TrainerParams:
    def __init__(self):
        self.T_train = 360
        self.T_val = 36
        self.testing_window = 1
        self.shrinkage_list = np.linspace(1e-12,10,50)
        self.save_ins = False

        # this is the number of individual saving chunks.
        # by this we mean the number of individual df contianing some oos performance that will be saved before merged.
        # too big and we risk loosing some processing, too small and we will make a mess of the merging process.
        self.nb_chunks = 25
        #
        self.min_nb_chunks_in_cluster = 10

# store all parameters into a single object
class Params:
    def __init__(self):
        self.name_detail = ''
        self.name = ''
        self.seed = 12345
        self.train = TrainerParams()
        self.rf = RandomFeaturesParams()

        self.update_model_name()

    def update_model_name(self):
        # function to overwrite
        self.name = self.name_detail

    def print_values(self):
        """
        Print all parameters used in the model
        """
        for key, v in self.__dict__.items():
            try:
                print('########', key, '########')
                for key2, vv in v.__dict__.items():
                    print(key2, ':', vv)
            except:
                print(v)

    def update_param_grid(self, grid_list, id_comb, verbose = False):
        ind = []
        for l in grid_list:
            t = np.arange(0, len(l[2]))
            ind.append(t.tolist())
        combs = list(itertools.product(*ind))
        if verbose:
            print('comb', str(id_comb + 1), '/', str(len(combs)))
        c = combs[id_comb]

        for i, l in enumerate(grid_list):
            self.__dict__[l[0]].__dict__[l[1]] = l[2][c[i]]

    def save(self, save_dir, file_name='/parameters.p'):
        # simple save function that allows loading of deprecated parameters object
        # first we create the dir if it doesn't exists already
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)


        df = pd.DataFrame(columns=['key', 'value'])

        for key, v in self.__dict__.items():
            try:
                for key2, vv in v.__dict__.items():
                    temp = pd.DataFrame(data=[str(key) + '_' + str(key2), vv], index=['key', 'value']).T
                    df = df.append(temp)

            except:
                temp = pd.DataFrame(data=[key, v], index=['key', 'value']).T
                df = pd.concat([df,temp],axis=0)
        df.to_pickle(save_dir + file_name)

    def load(self, load_dir, file_name='/parameters.p'):
        # simple load function that allows loading of deprecated parameters object
        df = pd.read_pickle(load_dir + file_name)
        # First check if this is an old pickle version, if so transform it into a df
        if type(df) != pd.DataFrame:
            loaded_par = df
            df = pd.DataFrame(columns=['key', 'value'])
            for key, v in loaded_par.__dict__.items():
                try:
                    for key2, vv in v.__dict__.items():
                        temp = pd.DataFrame(data=[str(key) + '_' + str(key2), vv], index=['key', 'value']).T
                        df = df.append(temp)

                except:
                    temp = pd.DataFrame(data=[key, v], index=['key', 'value']).T
                    df = df.append(temp)

        no_old_version_bug = True

        for key, v in self.__dict__.items():
            try:
                for key2, vv in v.__dict__.items():
                    t = df.loc[df['key'] == str(key) + '_' + str(key2), 'value']
                    if t.shape[0] == 1:
                        tt = t.values[0]
                        self.__dict__[key].__dict__[key2] = tt
                    else:
                        if no_old_version_bug:
                            no_old_version_bug = False
                            print('#### Loaded parameters object is depreceated, default version will be used')
                        print('Parameter', str(key) + '.' + str(key2), 'not found, using default: ',
                              self.__dict__[key].__dict__[key2])

            except:
                t = df.loc[df['key'] == str(key), 'value']
                if t.shape[0] == 1:
                    tt = t.values[0]
                    self.__dict__[key] = tt
                else:
                    if no_old_version_bug:
                        no_old_version_bug = False
                        print('#### Loaded parameters object is depreceated, default version will be used')
                    print('Parameter', str(key), 'not found, using default: ', self.__dict__[key])





import pandas as pd
import numpy as np
import statsmodels.api as sm
from enum import Enum
import os
import didipack


class LatexPaper:

    def __init__(self, dir_, fig_type_default='.png'):
        if dir_[-1] != '/':
            dir_ += '/'
        self.dir = dir_
        self.dir_tables = dir_ + 'tables/'
        self.dir_figs = dir_ + 'figs/'
        self.dir_sec = dir_ + 'sec/'
        self.fig_type_default = fig_type_default

    def create_paper(self, tex_name, title="Paper Title", author="Firstname Name", sec=[]):
        tex_name = tex_name.replace('.tex','')
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)
        if not os.path.exists(self.dir_sec):
            os.makedirs(self.dir_sec)
        if not os.path.exists(self.dir_figs):
            os.makedirs(self.dir_figs)
        if not os.path.exists(self.dir_tables):
            os.makedirs(self.dir_tables)

        with open(f'{self.dir}versionPO.sty', 'w') as txt:
            txt.write(didipack.TexRessources.versionPO_sty)

        if not os.path.exists(f'{self.dir_sec}abstract.tex'):
            with open(f'{self.dir_sec}abstract.tex', 'w') as txt:
                txt.write('')

        for r in sec:
            if not os.path.exists(f'{self.dir_sec + r}.tex'):
                with open(f'{self.dir_sec + r}.tex', 'w') as txt:
                    txt.write('')

        if not os.path.exists(f'{self.dir}/bib.bib'):
            with open(f'{self.dir}/sec/bib.bib', 'w') as txt:
                txt.write('')

        with open(f'{self.dir}jf.sty', 'w') as txt:
            txt.write(didipack.TexRessources.jf_sty)

        with open(f'{self.dir + tex_name}.tex', 'w') as txt:
            txt.write(didipack.TexRessources.m_tex_1 + title)
            txt.write(didipack.TexRessources.m_tex_2 + author)
            txt.write(didipack.TexRessources.m_tex_3)
            for r in sec:
                t = r"%%%%%%%%%%%%%%%%%%%%%%%%%" + '\n' \
                                                   r"\section{" + r + "}" + '\n' \
                                                                            r"\label{sec:" + r + "}" + '\n' \
                                                                                                       r"%%%%%%%%%%%%%%%%%%%%%%%%" + '\n' \
                                                                                                                                     r"\input{sec/" + r + ".tex}" + '\n' + '\n'

                txt.write(t)

            txt.write(didipack.TexRessources.m_tex_4 + "bib")
            txt.write(didipack.TexRessources.m_tex_5)

    def append_table_to_sec(self, table_name, sec_name, caption='caption',overall_caption=None,sub_dir=None, resize=-1,position='b'):
        table_name = table_name.replace('.tex', '')
        sec_name = sec_name.replace('.tex', '')
        if overall_caption is None:
            overall_caption = table_name
        t0 = "\n"+ '\\begin{table}['+position+'!] \n \\centering \n \\caption{'
        t1 = '}\n\\label{table:'
        if resize > 0:
            t2 = f'}}\n \\resizebox{{{resize}\\textwidth}}{{!}}{{\\input{{tables/'
            t3 = '.tex}} \n\\end{table}'
        else:
            t2 = '}\n \\input{tables/'
            t3 = '.tex} \n\\end{table}' + '\n'

        if sub_dir is None:
            final_name = table_name
        else:
            final_name = sub_dir+'/'+table_name

        f = t0 + caption + t1 + overall_caption + t2 + final_name + t3

        with open(f'{self.dir_sec + sec_name}.tex', 'a') as txt:
            txt.write(f)

    fig_names = ['f1', 'f2.png'];
    sec_name = 'some_sec';
    main_caption = 'CAPTION';
    fig_labels = ['a', 'b'];
    fig_type_default = '.png'

    def append_fig_to_sec(self, fig_names, sec_name, main_caption='CAPTION', size=None, fig_captions=[], fig_type=None,overall_label=None, sub_dir = None, position = 't'):
        if fig_type is None:
            fig_type = self.fig_type_default

        if not type(fig_names) == list:
            fig_names = [fig_names]

        fig_names = [x.replace(fig_type, '') for x in fig_names]


        if overall_label is None:
            overall_label = fig_names[0]

        if size is None:
            # select default size
            if len(fig_names) == 1:
                size = "80mm"
            if len(fig_names) in [2, 4, 8]:
                size = r"0.45\linewidth"
            if len(fig_names) in [3, 6, 9]:
                size = r"0.3\linewidth"
            if size is None:
                size = r"0.45\linewidth"

        t1 = "\n"+rf"\begin{{figure}}[{position}!] " + '\n' + "\centering"
        t2 = ''
        for i in range(len(fig_names)):
            let = "abcdefghijklmnopqrstuvwxyz"[i]

            if i < len(fig_captions):
                cap = fig_captions[i]
            else:
                cap = ''
            if sub_dir is None:
                name = fig_names[i]
            else:
                name = sub_dir +'/' + fig_names[i]

            t2 += r"\subfigure[" + cap + r"]{\label{fig:" + overall_label + "_" + let + r"}\includegraphics[width=" + size + r"]{figs/" + name + fig_type + r"}}" + '\n'
        t3 = r"\caption{" + main_caption + r"}" + '\n' +r"\label{fig:" + overall_label + r"}" +'\n'+ r"\end{figure}" + '\n'
        f = t1 + t2 + t3

        with open(f'{self.dir_sec + sec_name}.tex', 'a') as txt:
            txt.write(f)

    def append_text_to_sec(self, sec_name, text):
        txt = open(f'{self.dir_sec + sec_name}.tex','a')
        txt.write("\n"+text+"\n")
        txt.close()

    def create_new_sec(self, sec_name, text=""):
        with open(f'{self.dir_sec + sec_name}.tex', 'w') as txt:
            txt.write(text)





import sys
import pandas as pd

def convert_to_int(value):
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        if value.isdigit():
            return int(value)
    return value


class Argument(pd.Series):
    def get_value(self, v: str):
        if v in self.index:
            return self[v]
        else:
            return 0

    def __getattr__(self, name: str):
        return self.get_value(name)

def parse():
    # Default values for positional arguments
    first_pos = 0
    second_pos = 0
    third_pos = 0

    # Dictionary to hold named arguments with default value 0
    named_args = {}

    # List to hold remaining positional arguments
    positional_args = []

    # Parse through command-line arguments
    iterator = iter(sys.argv[1:])
    for arg in iterator:
        if arg.startswith('--'):
            # Remove leading '--' and split at the '=' to get key and value
            key, value = arg[2:].split('=')
            named_args[key] = convert_to_int(value)
        else:
            positional_args.append(arg)

    # Fill in positional arguments if they exist
    if len(positional_args) >= 1:
        first_pos = convert_to_int(positional_args[0])
    if len(positional_args) >= 2:
        second_pos = convert_to_int(positional_args[1])
    if len(positional_args) >= 3:
        third_pos = convert_to_int(positional_args[2])

    # Combine positional and named arguments in one dictionary
    all_args = {'a': first_pos, 'b': second_pos, 'c': third_pos}
    all_args.update(named_args)

    # Create an arguemnt to output
    args = Argument(all_args)

    return args

if __name__ == "__main__":
    result = parse()



import os
import sys
import numpy as np
import pandas as pd
import tqdm
from didipack.parameters import *
from matplotlib import pyplot as plt
from typing import Tuple, List
from didipack.utils_didi.general import General
from didipack.trainer.trainer_ridge import TrainerRidge


def generate_fake_data(N: int = 1000, P: int = 100) -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
    """
    Generates a dataset of fake data for testing purposes.

    Args:
        N (int): The number of rows in the generated data.
        P (int): The number of columns in the generated data.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]: A tuple containing the fake data, its mean along the columns, a series representing dates, and a series representing ids.
    """
    fake_data = pd.DataFrame(np.random.normal(size=(N, P)))
    y = pd.DataFrame(fake_data.mean(1))
    dates = pd.Series(np.arange(N))
    ids = dates.copy()

    return fake_data, y, dates, ids


def create_t_ind_index_from_dates_series(dates):
    return dates.drop_duplicates().sort_values().reset_index(drop=True).reset_index().rename(columns={'index':'t_ind'})


def get_start_dates(dates: pd.Series, T_train, testing_window) -> List:
    """
    Gets start dates for training.

    Args:
        dates (pd.Series): A series containing the dates for the data.
        T_train (int): The length of the training period.  IF T_TRAIN is NEGATIVE, we have an expanding window, minimum value T_Train
        testing_window (int): The length of the testing period.

    Returns:
        List[int]: A list containing the start dates for training.
    """

    start_dates = [dates.min() + np.abs(T_train)]
    while max(start_dates) < max(dates):
        start_dates.append(max(start_dates) + testing_window)

    return start_dates


def get_chunks(start_dates: List[int], nb_chunks: int, temp_save_dir: str) -> Tuple[
    List[Tuple[str, np.ndarray]], List[Tuple[str, np.ndarray]]]:
    """
    Gets chunks of start dates to process.

    Args:
        start_dates (List[int]): A list containing the start dates for training.
        nb_chunks (int): The number of chunks to split the data into.
        temp_save_dir (str): The directory where temporary data is saved.

    Returns:
        Tuple[List[Tuple[str, np.ndarray]], List[Tuple[str, np.ndarray]]]: A tuple containing two lists, one for the chunks that still need to be processed and one for the chunks that have already been processed.
    """
    chunks_of_start_dates = np.array_split(start_dates, nb_chunks)
    chunks_of_start_dates = [(f'chunk_{i}.p', x) for i, x in enumerate(chunks_of_start_dates)]
    processed_chunks = os.listdir(temp_save_dir)

    chunks_still_to_process = []
    chunks_already_processed = []
    for chunk in chunks_of_start_dates:
        if chunk[0] in processed_chunks:
            chunks_already_processed.append(chunk)
        else:
            chunks_still_to_process.append(chunk)

    return chunks_still_to_process, chunks_already_processed


def get_tasks_for_current_node(chunks_still_to_process: List[Tuple[str, np.ndarray]], nb_nodes_total: int,
                               min_nb_chunks_in_cluster: int, grid_id: int) -> List[Tuple[str, np.ndarray]]:
    """
    Gets tasks for the current node.

    Args:
        chunks_still_to_process (List[Tuple[str, np.ndarray]]): A list containing the chunks that still need to be processed.
        nb_nodes_total (int): The total number of nodes.
        min_nb_chunks_in_cluster (int): The minimum number of chunks in a cluster.
        grid_id (int): The id of the current grid.

    Returns:
        List[Tuple[str, np.ndarray]]: A list containing the tasks for the current node.
    """
    to_run_across_nodes = np.array_split(chunks_still_to_process, nb_nodes_total)
    if max([len(x) for x in to_run_across_nodes]):
        to_run_across_nodes = list(General.smart_chunks(chunks_still_to_process, min_nb_chunks_in_cluster))
        print(f'Splitting in a smart way {[len(x) for x in to_run_across_nodes]}')
    # [len(x) for x in to_run_across_nodes]
    to_run_now = to_run_across_nodes[grid_id]

    return to_run_now






if __name__ == "__main__":
    try:
        grid_id = int(sys.argv[1])
        model_id = int(sys.argv[2])
        print('Running with args', grid_id, flush=True)
    except:
        print('Debug mode on local machine', flush=True)
        grid_id = 0
        model_id = 5

    par = Params()
    par.train.testing_window = 5
    trainer = TrainerRidge(par)
    nb_nodes_total = 5
    verbose = True
    temp_save_dir = './res/temp2/'
    os.makedirs(temp_save_dir, exist_ok=True)

    fake_data, y, dates, ids = generate_fake_data()


    start_dates = get_start_dates(dates, par.train.T_train, par.train.testing_window)
    chunks_still_to_process, chunks_already_processed = get_chunks(start_dates,  par.train.nb_chunks, temp_save_dir)
    if verbose:
        print(f'Already processed {len(chunks_already_processed)}, Still to processed {len(chunks_still_to_process)}')

    to_run_now = get_tasks_for_current_node(chunks_still_to_process, nb_nodes_total, par.train.min_nb_chunks_in_cluster, grid_id)


    # running the analysis
    k = 0
    for chunk in to_run_now:
        k+=1
        df_oos_pred = pd.DataFrame()
        for start_id in tqdm.tqdm(chunk[1],f'Chunks {k} ({chunk[0]}) out of {len(to_run_now)}'):
        # for start_id in chunk[1]:
            y_test, _ = trainer.train_at_time_t(x=fake_data, y=y, ids=ids, times=dates, t_index_for_split=start_id)
            df_oos_pred = pd.concat([df_oos_pred, y_test], axis=0)
        df_oos_pred.to_pickle(temp_save_dir + chunk[0])



from matplotlib import pyplot as plt
import matplotlib.pylab as pylab
import pandas as pd

params = {'axes.labelsize': 14,
          'axes.labelweight': 'bold',
          'xtick.labelsize': 12,
          'ytick.labelsize': 12,
          'axes.titlesize': 12}
pylab.rcParams.update(params)

pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


class DidiPlot:
    LINE_STYLE = [
        '-', '--', ':', '-.',(0,(1,10)),(0,(5,19))
    ]
    # COLOR = [
    #     'black', 'gray', 'black', 'dimgray'
    # ]

    COLOR = [
        'black', 'blue', 'green', 'red','purple','brown'
    ]


import os
import pandas as pd
import numpy as np
from typing import Optional, Tuple
import time
class RandomFeaturesSpecs:
    def __init__(
            self,
            distribution = 'normal',
            distribution_parameters: list =None,
            activation = 'cos_and_sin',
            bias_distribution:str = None,
            bias_distribution_parameters: list =None,
            ranking: bool = True,
            binning_feature: bool = False,
            random_rotation: bool = False
    ):
        # if distribution =='normal':
        #     assert (type(distribution_parameters)==list) & (len(distribution_parameters)==2), \
        #         "For a 'normal' distribution of parameters please define distribution_parameters as a list [mean, std]"
        # if bias_distribution =='normal':
        #     assert (type(bias_distribution_parameters)==list) & (len(bias_distribution_parameters)==2), \
        #         "For a 'normal' distribution of parameters please define distribution_parameters as a list [mean, std]"
        if distribution_parameters is None:
            distribution_parameters = [0,1]

        self.distribution = distribution
        self.distribution_parameters = distribution_parameters
        self.activation = activation
        self.bias_distribution = bias_distribution
        self.bias_distribution_parameters = bias_distribution_parameters
        self.number_features = None
        self.binning_feature = binning_feature
        self.random_rotation = random_rotation
        self.ranking = ranking


class RandomFeaturesGenerator:
    """



    """

    # the next shows the number of basic_parameters defining the distribution
    distribution_requirements = {
        "beta": 2,
        "binomial": 2,
        "chisquare": 0,
        "dirichlet": 1,
        "exponential": 0,
        "f": 2,
        "gamma": 1,
        "geometric": 1,
        "gumbel": 2,
        "hypergeometric": 2,
        "laplace": 2,
        "logistic": 2,
        "lognormal": 2,
        "logseries": 1,
        "multinomial": 2,
        "multivariate_normal": 2,
        "negative_binomial": 2,
        "noncentral_chisquare": 2,
        "noncentral_f": 3,
        "normal": 2,
        "normal_as_old": 2,
        "pareto": 1,
        "poisson": 1,
        "power": 1,
        "rayleigh": 1,
        "standard_cauchy": 0,
        "standard_exponential": 0,
        "standard_gamma": 1,
        "standard_normal": 0,
        "standard_t": 1,
        "triangular": 3,
        "uniform": 2,
        "vonmises": 2,
        "wald": 2,
        "weibull": 1,
        "zipf": 1,
        "gaussian_mixture": 0,
    }

    permitted_activation_functions = [
        "cos",
        "sin",
        "exp",
        "arctan",
        "tanh",
        "ReLu",
        "Elu",
        "SoftPlus",
        "cos_and_sin",
    ]


    def __init__(
            self,
            build_factor: bool = True,
            rf_spec_list: [list, RandomFeaturesSpecs] = RandomFeaturesSpecs(),
            compute_instrument_quantities: object = False,
            half_linear_features: bool = False

    ):
        if type(rf_spec_list) == RandomFeaturesSpecs:
            rf_spec_list = [rf_spec_list]

        self.build_factors = build_factor
        self.rf_spec_list = rf_spec_list
        self.compute_instrument_quantities = compute_instrument_quantities
        self.half_linear_features = half_linear_features



    def generate_random_features_from_list_with_potential_ranking(
            self,
            seed: int,
            batch_nb_rf: int,
            date_ids: np.ndarray,
            x_mat: np.ndarray,
            y_mat: [np.ndarray, None] = None,
            permute_randomly: bool = False,
    ) -> pd.DataFrame:

        """
        This is a key function for generating random features, with the possibility to cross-sectionally rank them
        If msrr or factor_msrr, then we will actually multiply signals * returns
        Parameters
        ----------
        seed :
        compute_instrument_quantities: if true, we are NOT going to produce the random factors or even feature,
        only use random signals to compute buidling blocks of pricing errors
        msrr :
        factor_msrr: if true, then we actually build factors
        batch_nb_rf : number random features
        pre_specified_list_of_specs_for_random_features :
        date_ids: We use then to groupby when we build factors
        y_mat: (NT) \times 1 panel of stock returns
        x_mat :

        Returns
        -------
        first we generate random_features: (NT) \times (number_features_in_subset)
        matrix: array of random features for the panel of
        N stocks and T dates

        If factor_msrr = True, we return (T \times number_features_in_subset) because we groupby dates !
        If factor_msrr = False, then we return original size  (NT) \times (number_features_in_subset)


        """

        # we define here the number of random features per group.
        size_per_group = int(np.floor(batch_nb_rf/len(self.rf_spec_list)))
        left_over = int(batch_nb_rf-size_per_group*len(self.rf_spec_list))
        for i in range(len(self.rf_spec_list)):
            if i == len(self.rf_spec_list)-1:
                self.rf_spec_list[i].number_features = size_per_group+left_over
            else:
                self.rf_spec_list[i].number_features = size_per_group


        random_features_for_generation = []
        for spec in self.rf_spec_list:
            random_features_for_generation.append(
                self._generate_random_features_from_spec(
                    seed=seed,
                    spec=spec,
                    x_mat=x_mat,
                    date_ids=date_ids
                )
            )
        random_features_for_generation = np.concatenate(random_features_for_generation,axis=1)
        if self.build_factors:
            # note that we update the dates according to the groupby so that we can get the updated dates as output
            random_features_for_generation = self._build_random_factors(x_mat=random_features_for_generation, y_mat=y_mat, dates_ids=date_ids)
        else:
            random_features_for_generation = pd.DataFrame(random_features_for_generation,index=date_ids.flatten())
        random_features_for_generation.index.name = 'date'

        if permute_randomly:
            perm = np.random.permutation(random_features_for_generation.shape[1])
            random_features_for_generation = random_features_for_generation.loc[:, perm]

        return random_features_for_generation

    def _rank_features_cross_sectionally(self,
                                         unkranked_x: np.ndarray,
                                         date_ids: np.ndarray,
                                         print_time: bool = False):
        """
        Ranking features on each date. This is done in numpy for speed, but the drawback is that
        we assume dates are pre-ordered
        Parameters
        ----------
        unkranked_x
        date_ids
        ranking
        axis_
        print_time : If true we print the run tiem
        use_pands :  IF true we run the pandas version.

        Returns
        -------

        """

        date_ids = np.array(date_ids)

        """
        Important: the code assumes that date_ids are already sorted! 
        Pre-process the data so that date_ids are increasing !! 
        """
        # here it is important that dates are sorted because when we concatenate,
        # we do so in the order of dates
        start = time.time()
        ranked_signals = pd.DataFrame(unkranked_x)
        if len(date_ids.shape)==2:
            date_ids = date_ids.flatten()
        ranked_signals = ranked_signals.groupby(date_ids).rank(pct=True, axis=0) - 0.5
        ranked_signals = ranked_signals.values
        if print_time:
            print('Time pandas', time.time() - start)

        ranked_signals[np.isnan(unkranked_x)] = np.nan

        return ranked_signals

    @staticmethod
    def _check_distribution_requirements(
            distribution: str,
            distribution_parameters: list
    ):

        if distribution == "gaussian_mixture":
            return
        if distribution not in RandomFeaturesGenerator.distribution_requirements:
            raise Exception(
                f"{distribution} is not permitted. If you need it, do not be lazy and update the class"
            )
        elif (
                len(distribution_parameters)
                != RandomFeaturesGenerator.distribution_requirements[distribution]
        ):
            raise Exception(
                f"{distribution} "
                f"requires {RandomFeaturesGenerator.distribution_requirements[distribution]} basic_parameters"
            )


    def _build_random_factors(
            self,
            x_mat,
            y_mat: np.ndarray,
            dates_ids: np.ndarray) -> pd.DataFrame:
        """

        Parameters
        ----------
        x_mat: (NT) \times (number_features_in_subset)  matrix: array of random features for the panel of
        N stocks and T dates
        y_train: (NT) \times 1 panel of stock returns
        factor_msrr: if true, then we actually build factors
        dates_ids: dates. We use then to groupby when we build factors

        Returns
        -------
        If factor_msrr = True, we return (T \times number_features_in_subset) because we groupby dates !
        If factor_msrr = False, then we return original size  (NT) \times (number_features_in_subset)

        """
        x_mat = x_mat * y_mat.reshape(-1, 1)  # these are actually managed returns, R * F(S)
        if type(x_mat) == np.ndarray:
            x_mat = pd.DataFrame(x_mat)
        if len(dates_ids.shape)== 2:
            dates_ids = dates_ids.flatten()
        x_mat = x_mat.groupby(dates_ids).mean()
        return x_mat

    def _generate_random_features_from_spec(
            self,
            x_mat: np.ndarray,
            spec: RandomFeaturesSpecs,
            date_ids: np.ndarray,
            seed: int = None
    ) -> np.ndarray:

        """
        The starting points
        given a list of different specifications, generate random features for each of them
        normally, it should receive specifications as pre_specified_list_of_specs
        but if not, then we give specification details explicitly
        Parameters
        ----------
        number_features_in_subset: number of features per gamma
        x_mat: actual raw features
        reproducible, yet each chuck must yield new features. increment_seed takes care of it
        single_precision: this is for memory saving and gpu
        pre_specified_list_of_specs: the main thing: a list specifying all details of random features
        seed: random seed

        Returns
        -------

        """
        np.random.seed(seed)
        if spec.binning_feature:
            random_features_for_spec = self._generate_random_binning_features(x_mat=x_mat, spec=spec, seed=seed)
        else:
            random_features_for_spec = self.generate_random_neuron_features(x_mat=x_mat, spec=spec,seed=seed)
        if spec.ranking:
            random_features_for_spec = self._rank_features_cross_sectionally(
                unkranked_x= random_features_for_spec,
                date_ids=date_ids,
            )

        return random_features_for_spec


    @staticmethod
    def _apply_activation_to_multiplied_signals(
            multiplied_signals,
            activation: str
    ) -> np.ndarray:
        """
        this method takes as input signaled already multipled by some weights + cosntant: w*x+b
        and returns act(w*x+b)
        :rtype: object
        """
        if activation in ["cos", "sin", "exp", "arctan", "tanh"]:
            final_random_features = getattr(np, activation)(multiplied_signals)
        elif activation == "cos_and_sin":
            final_random_features = np.concatenate(
                [np.cos(multiplied_signals), np.sin(multiplied_signals)], axis=0
            )

        elif isinstance(activation, str) and activation.lower() == "relu":
            final_random_features = multiplied_signals * (multiplied_signals > 0)
        elif isinstance(activation, list) and activation[0].lower() == "elu":
            final_random_features = (
                                            multiplied_signals * (multiplied_signals > 0)
                                    ) + activation[1] * (np.exp(multiplied_signals) - 1) * (
                                            multiplied_signals < 0
                                    )
        elif activation.lower() == "softplus":
            final_random_features = np.log(1 + np.exp(multiplied_signals))
        elif activation.lower() == "linear":

            final_random_features = multiplied_signals
        else:
            raise Exception(f"activation function={activation} is not yet supported")
        return final_random_features

    def _add_bias(
            self,
            multiplied_signals,
            bias_distribution,
            bias_distribution_parameters,
            seed=0,
    ):
        """
        Careful, multiplied signals are assumed to be P \times n where P is the sumber of signals
        and n the number of observations
        Parameters
        ----------
        multiplied_signals :
        bias_distribution :
        bias_distribution_parameters :
        Returns
        -------
        """
        np.random.seed(seed)
        number_features = multiplied_signals.shape[0]
        random_bias = getattr(np.random, bias_distribution)(
            *bias_distribution_parameters, [number_features, 1]
        )
        # here we are using numpy broadcasting to add the same bias every time period
        multiplied_signals += random_bias
        return multiplied_signals

    def generate_random_neuron_features(
            self,
            x_mat: np.ndarray,
            seed: int,
            spec : RandomFeaturesSpecs
    ) -> np.ndarray:

        """

        this function builds random neuron features f(w'S+bias) where w is
        a vector of random weights and f is an activation function, and bias is a random bias

        One important special case is a gaussian mixture, whereby we generate random scales (gamma) from the interval
        [gamma_min, gamma_max], uniformly

        :param distribution_parameters:
        :param distribution:
        :param activation:
        :param number_features:
        :param bias_distribution:
        :param gamma_method: random or linear to randomly draw it or get a grid
        :param gama_param: if not none we use gamma and we draw it from x to y -> [x,y]
        :return:
        """
        np.random.seed(seed)
        signals = x_mat

        RandomFeaturesGenerator._check_distribution_requirements(
            spec.distribution, spec.distribution_parameters
        )
        if spec.bias_distribution:
            RandomFeaturesGenerator._check_distribution_requirements(
                spec.bias_distribution,
                spec.bias_distribution_parameters,
            )

        number_signals = signals.shape[1]
        size = [number_signals, spec.number_features]
        if spec.activation == "cos_and_sin":
            size = [number_signals, int(spec.number_features / 2)]
        # first we initialize the random seed

        # X = np.random.normal(0, a) means X is distributed as Normal(0, a^2).  (a=standard deviation)
        # This is an important property of Gaussian distributions: multiplying by a constant keeps is Gaussian,
        # just scales the standard deviation
        if spec.distribution != "gaussian_mixture":
            random_vectors = getattr(np.random, spec.distribution)(*spec.distribution_parameters, size)
        else:
            random_vectors = getattr(np.random, "standard_normal")(size)
            gamma_values = spec.distribution_parameters
            minimal_gamma = gamma_values[0]
            maximal_gamma = gamma_values[1]
            all_gamma_values = np.random.uniform(minimal_gamma, maximal_gamma, [1, size[1]])
            # now we use numpy broadcasting to do elemen-wise multiplication.
            random_vectors = random_vectors * all_gamma_values

        # w'x, where w is our random vector
        multiplied_signals = np.matmul(random_vectors.T, signals.T)
        if spec.bias_distribution:
            multiplied_signals = RandomFeaturesGenerator._add_bias(
                multiplied_signals, spec.bias_distribution, spec.bias_distribution_parameters
            )
        final_random_features = (
            RandomFeaturesGenerator._apply_activation_to_multiplied_signals(
                multiplied_signals, spec.activation
            )
        )

        return final_random_features.T

    @staticmethod
    def _generate_random_binning_features(x_mat: np.ndarray,
                                          spec:RandomFeaturesSpecs,
                                          random_seed=0,
                                          seed=0):
        """
           WARNING: THE FEATURES ARE SUPPOSED TO BE NORMALIZED!!!
           ALWAYS PRE-PROCESS THE DATA (USING ROLLING WINDOW) !!!
           signals are assumed to be T \times M
           :param random_seed:
           :param random_rotation:
           :param distribution_parameters:
           :param distribution:
           :param distribution_requirements:
           :param x_mat:
           :param number_features:
           :return:
           """
        np.random.seed(seed)
        RandomFeaturesGenerator._check_distribution_requirements(spec.distribution,
                                                                 spec.distribution_parameters)

        number_signals = x_mat.shape[1]
        size = [number_signals, spec.number_features]
        np.random.seed(random_seed)
        if spec.random_rotation:
            rotate = np.random.randn(x_mat.shape[1], x_mat.shape[1])
            tmp = np.matmul(rotate, rotate.T)
            _, eigvec = np.linalg.eigh(tmp)
            # now, eigenvectors give a random rotation
            signals_rotated = np.matmul(eigvec.T, x_mat.T).T
        else:
            signals_rotated = x_mat.copy()
        delta = getattr(np.random, spec.distribution)(*spec.distribution_parameters, size)
        delta = delta * (np.abs(delta) > (10 ** (- 10))) + (np.abs(delta) < (10 ** (- 10))) * (10 ** (-10))  # clip
        u_ = np.random.uniform(0, 1, [number_signals, spec.number_features]) * delta
        subtracted = signals_rotated.reshape(x_mat.shape[0], 1, number_signals) \
                     - u_.reshape(1, spec.number_features, number_signals)
        subtracted_and_divided = subtracted / delta.reshape(1, spec.number_features, number_signals)

        binned_signals = np.floor(subtracted_and_divided).reshape([x_mat.shape[0],
                                                                   x_mat.shape[1] * spec.number_features])
        return binned_signals

    @staticmethod
    def _generate_a_random_covariance_matrix_with_given_eigenvalues(
            eigenvalues: np.ndarray, seed=0
    ) -> np.ndarray:
        """

        Parameters
        ----------
        eigenvalues :

        Returns
        -------
        a random matrix with prescribed eigenvalues
        """
        eigenvalues = eigenvalues.flatten()
        size = len(eigenvalues)
        random_shocks = np.random.normal(size=[size, 2 * size])
        sigma = random_shocks @ random_shocks.T
        _, eigenvectors = np.linalg.eigh(sigma)
        # t1 = time.time()
        # eigenvectors = ortho_group.rvs(size, random_state=seed)
        # matrix = eigenvectors @ (np.diag(eigenvalues) @ eigenvectors.T)
        # t2 = time.time()
        # print(f'got the random O matrix in {t2 - t1}')
        matrix = eigenvectors @ (np.diag(eigenvalues) @ eigenvectors.T)
        return matrix




if __name__ == "__main__":
    pass
    # c_sigma = 3.0
    # list_of_widths = [300, 300, 600]
    #
    # X_train = np.random.normal(size=(100, 100))

    # distribution = "gaussian_mixture"
    # gamma = [1.0, 1.0]
    # activation = "relu"
    #
    # random_features = RandomFeaturesGenerator.generate_random_deep_neuron_features(
    #     c_sigma, list_of_widths, X_train, distribution, gamma, activation
    # )




